import json
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import rc

def read_multiqc_fastp(mqc_fastp_json):
    '''Parses the json file generated by MultiQC to generate a tsv report for FastP results.
    '''
    mqc_fastp_json_file = mqc_fastp_json
    with open(mqc_fastp_json_file, 'r') as my_file:
        json_data = my_file.read()
    #json_obj = json.loads(json_data)
    #multiqc_fastp_data = json_obj['report_general_stats_data'][0]
    multiqc_fastp_data = json.loads(json_data)['report_general_stats_data'][0]
    multiqc_fastp_data_df = pd.DataFrame.from_dict(multiqc_fastp_data, 
                                                   orient='index')
    return(multiqc_fastp_data_df)

def summarize_filter_human_step(mqc_filter_human_list):
    '''Parses all the stats files included in mqc_filter_human_list and
    creates a single merged table report.
    
    Args:
        mqc_filter_human_list: FilePath to the file containing a list of stats files
    Returns:
        A pandas.DataFRame
    '''
    try:
        with open(mqc_filter_human_list, 'r') as my_file:
            print(mqc_filter_human_list)
            bfiltered_files = [f.strip() for f in my_file.readlines()]
    except (TypeError, ValueError, IOError, EOFError) as e:
        print(e)
    else:
        header = list()
        multi_bfiltered_data = dict()
        for ix, bfiltered_file in enumerate(bfiltered_files):
            unit_name = bfiltered_file.split('/')[-1].split('.')[0]
            with open(bfiltered_file) as f:
                lines = f.read().split('\n')[:-1]
                if ix==0:
                    header = lines[0].split('\t')
                    multi_bfiltered_data[unit_name] = lines[1].split('\t')
                else:
                    multi_bfiltered_data[unit_name] = lines[1].split('\t')
        header[0] = 'ref'
        header[1] = 'pct_unambiguousReads'
        header[3] = 'pct_ambiguousReads'

        multi_bfiltered_stats = pd.DataFrame.from_dict(multi_bfiltered_data, 
                                                       orient='index', columns=header)
        return(multi_bfiltered_stats)

def summarize_preqc(mqc_trimadapters_fastps_json, mqc_filter_human_stats, mqc_trim3end_fastps_json):
    """Summarizes all the steps in the preQC workflow.
    
    Args:
        mqc_trimadapters_fastps_json: FilePath to the tsv report from step trimadapters
        mqc_filter_human_stats: FilePath to the tsv report from step filter_human
        mqc_trim3end_fastps_json: FilePath to the tsv report from step trim3end
    Returns:
        A pandas.DataFRame
    """
    try:
        trim_adapters_mqc_data = read_multiqc_fastp(mqc_trimadapters_fastps_json)
        filter_human_mqc_data = pd.read_csv(mqc_filter_human_stats, sep='\t', header=0, index_col=0)
        trim3end_mqc_data = read_multiqc_fastp(mqc_trim3end_fastps_json)
    except (ValueError,TypeError) as e:
        print(e)
    else:
        trim_adapters_mqc_tmp = pd.DataFrame()
        trim_adapters_mqc_tmp['total_reads'] = trim_adapters_mqc_data['before_filtering_total_reads']
        trim_adapters_mqc_tmp['trim_adapters_disc_reads'] = trim_adapters_mqc_tmp['total_reads'] - \
                                                            trim_adapters_mqc_data['after_filtering_total_reads']

        filter_human_mqc_tmp = pd.DataFrame()
        filter_human_mqc_tmp['filter_human_disc_reads'] = filter_human_mqc_data['assignedReads'].astype('int') 

        trim3end_mqc_tmp = pd.DataFrame()
        trim3end_mqc_tmp['trim_3end_disc_reads'] = trim3end_mqc_data['before_filtering_total_reads'] - \
                                                   trim3end_mqc_data['after_filtering_total_reads']

        tmp_df = pd.concat([trim_adapters_mqc_tmp, filter_human_mqc_tmp, trim3end_mqc_tmp], axis=1, join='outer', sort=True)
        tmp_df.loc[trim3end_mqc_data.index,'dedupe_disc_reads'] = tmp_df.loc[trim3end_mqc_data.index,'total_reads'] - \
                                                                  tmp_df.loc[trim3end_mqc_data.index,'trim_adapters_disc_reads'] - \
                                                                  tmp_df.loc[trim3end_mqc_data.index,'filter_human_disc_reads'] - \
                                                                  trim3end_mqc_data['before_filtering_total_reads'].values
        tmp_df.loc[trim3end_mqc_data.index,'total_disc_reads'] = tmp_df.loc[trim3end_mqc_data.index,'total_reads'] - \
                                                                 trim3end_mqc_data['after_filtering_total_reads']

        tmp_df.loc[trim3end_mqc_data.index,'total_kept_reads'] = trim3end_mqc_data['after_filtering_total_reads']

        # Adding columns with percentages of discarded reads in each step
        tmp_df['trim_adapters_pct_disc'] = tmp_df['trim_adapters_disc_reads'] / tmp_df['total_reads']*100
        tmp_df['filter_human_pct_disc'] = tmp_df['filter_human_disc_reads'] / tmp_df['total_reads']*100
        tmp_df['dedupe_pct_disc'] = tmp_df['dedupe_disc_reads'] / tmp_df['total_reads']*100
        tmp_df['trim_3end_pct_disc'] = tmp_df['trim_3end_disc_reads'] / tmp_df['total_reads']*100
        tmp_df['total_pct_disc'] = tmp_df['total_disc_reads'] / tmp_df['total_reads']*100
        tmp_df['total_pct_kept'] = tmp_df['total_kept_reads'] / tmp_df['total_reads']*100

        # Re-order columns
        tmp_columns = ['total_reads','trim_adapters_disc_reads','filter_human_disc_reads','dedupe_disc_reads','trim_3end_disc_reads'] + \
                      tmp_df.columns.tolist()[5:]
        tmp_df = tmp_df[tmp_columns]

        # Add sample, run, lane columns
        sample_run_lane = pd.DataFrame.from_dict({unit:{'Sample':unit.split('-')[0], 
                                                        'Run':unit.split('-')[1].split('_')[0], 
                                                        'Lane':unit.split('-')[1].split('_')[1]
                                                       } 
                                                  for unit in tmp_df.index
                                                 }, 
                                                 orient='index'
                                                )

        sample_run_lane = pd.concat([sample_run_lane, tmp_df], axis=1, join='outer', sort=False)
        return(sample_run_lane)

def plot_preqc_summary(preqc_summary_df, by='units', plot_type='raw'):
    """Creates a radial bar chart from the dataframe generted by summarize_preqc.
    
    Args:
        preqc_summary_df: pd.DataFrame object generated by summarize_preqc()
        by: either 'units' or 'samples'. it specifies how to aggregate the data.
        plot_type: either 'raw' or 'pct'. If 'pct' a percent stacked barchar is returned.
    Returns:
        matplotlib.figure.Figure
    """
    import inspect
    by_options = ['units','samples']
    type_options = ['raw','pct']
    
    frame = inspect.currentframe()
    args_type = ['pandas.DataFrame', by_options, type_options]
    UNDEFINED = object()
    def print_message(frame, args_type):
        args, _, _, values = inspect.getargvalues(frame)
        print('{}()'.format(inspect.getframeinfo(frame)[2]))
        print('Args:')
        for ix,arg in enumerate(args):
            print('    {}: {}'.format(arg, str(args_type[ix])))
    
    if (not isinstance(preqc_summary_df, pd.DataFrame)) or (by not in by_options) or (plot_type not in type_options):
        print('Invalid arguments')
        print_message(frame,args_type)
        return(UNDEFINED)
    else:
        try:
            if plot_type == 'raw':
                columns_to_plot = ['trim_adapters_disc_reads',
                                   'filter_human_disc_reads', 
                                   'dedupe_disc_reads',
                                   'trim_3end_disc_reads',
                                   'total_kept_reads']
                sample_run_lane = preqc_summary_df[['total_reads','Sample']+columns_to_plot].copy()
                if by == 'samples':
                    sample_run_lane = sample_run_lane.groupby(['Sample']).agg('sum')
                bottom_lim = 0.0
                top_lim = sample_run_lane['total_reads'].max()
                # This uses the given y-units of bottom spacing for each bar
                # You may increase this number a bit to have more spacing between bars and text.
                bottom_spacing = top_lim/1.8
                rgrids_positions = [sample_run_lane['total_reads'].max()/(i) for i in [10,4,2,1]] + bottom_spacing
                rgrids_positions_labels = [' {}M'.format(nr) 
                                           for ix,nr in enumerate(np.round((rgrids_positions-bottom_spacing)/1000000, decimals=3))]
            else:
                columns_to_plot = ['trim_adapters_pct_disc',
                                   'filter_human_pct_disc', 
                                   'dedupe_pct_disc',
                                   'trim_3end_pct_disc',
                                   'total_pct_kept']
                sample_run_lane = preqc_summary_df[['total_reads','Sample']+columns_to_plot].copy()
                if by == 'samples':
                    sample_run_lane = sample_run_lane.groupby(['Sample']).agg('mean')
                bottom_lim = 0.0
                top_lim = 100
                bottom_spacing = top_lim/1.8
                rgrids_values = [25,50,75,100]
                rgrids_positions =  np.add(rgrids_values,bottom_spacing)
                rgrids_positions_labels = [' {}%'.format(nr) for nr in rgrids_values]
                    
            sample_run_lane.sort_values(by=['total_reads'], inplace=True)
            # Colors for each column (nested bars inside each bar)
            barColors = ['#fdb863', '#e08214', '#b35806', '#542788','#8073ac'][::-1]
        except (TypeError, ValueError) as e:
            print(e)
        else:
            # This variable was used for testing purposes to see how fold increases 
            # of the datapoints affects the plot
            lenMult = 1
            # Nr of datapoints to plot
            nr_h_bars = len(sample_run_lane)*lenMult
            # This sets the number of x-positions left empty at the beginning
            # so that enough space is left for y-labels
            nr_empty_bars = np.int(np.ceil(nr_h_bars/40))
            empty_bars = [0]*nr_empty_bars
            empty_bars_labels = ['']*len(empty_bars)
            # re-set the number of x-positions (data points) to plot
            nr_h_bars = nr_h_bars+nr_empty_bars
    
            # The positions of the bars on the x-axis
            # Each position corresponds to a radian degree in a unit circunference (radius is 1)
            # The circumference (2pi) is divided into nr_h_bars sections
            r= np.arange(0,2*np.pi,2*np.pi/nr_h_bars) # for radial barchart 
            r= r[0:nr_h_bars]
            # Bar width
            barWidth = (2*np.pi)/(nr_h_bars)*0.9 # for radial barchart 
    
            # Define the columns to use for the stacked bars, 
            # values of each column are plot on top of one another
            barColumns = columns_to_plot[::-1]
    
            # This uses the given y-units of bottom spacing for each bar
            # You may increase this number a bit to have more spacing between bars and text.
            bottomBars = [bottom_spacing]*nr_h_bars
    
            # Lets set the figure size depending on the amount of data points to plot
            # Labels until 1000 datapoints are readable
            # Consider adding additional elif scenarios to increase size for larger datasets
            if nr_h_bars <=200:
                fig = plt.figure(figsize=(10, 10))
                font_size = 7.5
            elif nr_h_bars <= 500:
                fig = plt.figure(figsize=(20, 20))
                font_size = 'x-small'
            else:
                fig = plt.figure(figsize=(28, 28))
                font_size = 6

            # Here we do the actual plotting
            # ------------------------------
            # Values for each column for a datapoint (i.e sequencing unit, or sample)
            # are plot on top of one another
            ax = fig.add_axes([0.1,0.1,0.75,0.75], projection='polar')
            for ix,column in enumerate(barColumns):
                bar_values=empty_bars+sample_run_lane[column].tolist()*lenMult
                bars=ax.bar(r, bar_values, bottom=bottomBars, 
                            color=barColors[ix], edgecolor='white', width=barWidth, label=column)
                bottomBars = np.add(bottomBars,bar_values).tolist()
    
            # Here we add annotations (e.g labels, legend)
            # --------------------------------------------
            rotations = np.rad2deg(r)
            y0,y1 = ax.get_ylim()
            bar_labels= empty_bars_labels + sample_run_lane.index.tolist()*lenMult
            for x, y_pos, rotation, label in zip(r, bottomBars, rotations, bar_labels):
                offset = (y_pos)/(y1-y0)
                lab = ax.text(0, 0, label, fontsize=font_size, transform=None, ha='center', va='center', fontstretch='ultra-condensed')
                renderer = ax.figure.canvas.get_renderer()
                bbox = lab.get_window_extent(renderer=renderer)
                invb = ax.transData.inverted().transform([[0,0],[bbox.width,0] ])
                lab.set_position((x,offset+(invb[1][0]-invb[0][0])/2.*2.7) )
                lab.set_transform(ax.get_xaxis_transform())
                lab.set_rotation(rotation)
            #rgrids_positions = sample_run_lane['total_reads'].describe()[[3,5,7]].values + bottom_spacing
            #rgrids_positions_labels = [' {}M'.format(nr)
            #                           for ix,nr in enumerate(np.round((rgrids_positions-bottom_spacing)/1000000, decimals=3))]
            ax.set_rgrids(rgrids_positions,
                          labels=rgrids_positions_labels, 
                          angle=np.rad2deg(r[np.int(nr_empty_bars/2-1)]), 
                          family='serif', ha='left', va='center', weight='bold',fontstretch='ultra-condensed', fontsize=8.5)
            ax.legend(bbox_to_anchor=(0.5, 0.5), loc='center', frameon=False, title='MetaSnk: preQC')
            # A few extra edits for aesthetics
            # -------------------------------
            ax.xaxis.grid(False)
            ax.spines["polar"].set_visible(False)
            ax.set_xticklabels([])
    
            return(fig)